//vi: set ts=4
#pragma option INSTRUCTIONTIMING	// add instruction timing to listing file
#pragma option f0			// don't insert page breaks in the listing
#pragma	option NOINIT			// do not automatically provide pre-initialization
#pragma option RAMMAP			// display ram usage at the bottom of the listing

#include	"cy640.h"
#include 	"usb.h"
#include 	"main.h"

//********************************************************
//				PIN ASSIGNMENTS
//********************************************************

//	port 0
#define	LED				0x55	// Port0[6,4,2,0]

// port 1
#define	SDATA			0x01	// Port1[0]
#define	SCLK			0x02	// Port1[1]

// port 3
#define	S3DATA			0x02	// Port3[1]
#define	S3CLK			0x01	// Port3[0]

// port 2
#define	XRES			0x08	// Port2[3]

#define	HIZPU			0x30	// Port2[5:4]


//*********************************************************
// Constants
//*********************************************************

// STATE MACHINE CONSTANTS
// EP0 IN TRANSACTIONS

#define	EP0_IN_IDLE				0x00
#define	CONTROL_READ_DATA		0x02
#define NO_DATA_STATUS			0x04
#define	EP0_IN_STALL			0x06

// FLAG CONSTANTS
// EP0 NO-DATA CONTROL FLAGS

#define	ADDRESS_CHANGE_PENDING		0x00
#define	NO_CHANGE_PENDING			0x02

// RESPONSE SIZES
#define	DEVICE_STATUS_LENGTH		0x02
#define	DEVICE_CONFIG_LENGTH		0x01
#define	ENDPOINT_STALL_LENGTH		0x02
#define	INTERFACE_STATUS_LENGTH		0x02
#define	INTERFACE_ALTERNATE_LENGTH	0x01
#define	INTERFACE_PROTOCOL_LENGTH	0x01

#define	NO_EVENT_PENDING		0x00
#define	EVENT_PENDING			0x02

//***** TRANSACTION TYPES
#define	TRANS_NONE				0x00
#define	TRANS_CONTROL_READ		0x01
#define	TRANS_CONTROL_WRITE		0x02
#define	TRANS_NO_DATA_CONTROL	0x03


//***** endpoint 1 IN buffer states
#define IN_IDLE				_LEMPTY
#define IN_TRANSFER			~_LEMPTY

#define	MORE_RD_DATA		_DREADY
#define NO_RD_DATA			~_DREADY


void ep0DataReturn(Byte);

static void Tx8Bits();
static void TxOP();
static void Tx30();
static void TestController();
static void Acquire();
static void Pacquire(Byte);
static void PSOCreset();
static void ReleaseSdata();
static void ReleaseBoth();
static void I2CStart(void);
static void I2CStop(void);
static void I2CBit(Byte d);
static int I2CTx(void);
static void I2CRx(Byte nack);

void	i2cWrite(void);
void	i2cRead(void);

#define TC(A,D) addr = A; data = D; TestController();

// DATA MEMORY VARIABLES

Byte int_temp;			// interrupt temporary register 

Byte ep0_data_toggle;	// data toggle for INs on endpoint zero
Byte *far data_start;	// address of request response data, as an offset
Byte data_count;		// number of bytes to send back to the host
Byte maximum_data_count;// request response size

Byte suspend_count;		// counter for suspend/resume
Byte suspend_flag;		// flag for bus suspend condition

Byte ep0_in_machine;	// endpoint0 state machine          
Byte ep0_in_flag;		// endpoint0 IN flag    
Byte ep0_transtype;		// endpoint0 tranaaction type   
Byte pending_data;		// no-data control transaction data         
Byte configuration;		// device config/unconfig status            

Byte ep1_data_toggle;	// data toggle for INs on endpoint one
Byte ep1_stall;			// endpoint 1 stall status  

Byte ep2_data_toggle;	// endpoint2 data toggle    
Byte ep2_stall;			// endpoint 2 stall status  

Byte ep0_dmabuff[8] @ 0xB8;
Byte ep1_dmabuff[32] @ 0xC0;
Byte ep2_dmabuff[32] @ 0xE0;

Byte suspend_ena;

unsigned int  Clock;

Byte in_state;
Byte in_timer;
Byte in_timeout_value;
Byte out_state;

Byte addr;
Byte cmd;
Byte mode;		// AddressSpace mode into PLD
Byte data;
Byte count;
Byte tcData;
Byte tcOp;
Byte programMode;
Byte sram_address;
Byte bulkMode;
Byte delay1;
Byte i,j;
Byte blockId;
Byte sequenceMode;

void
_error()
{
	STOP();
}


//******************************************************
//
//  Interrupt: Main
//  Purpose: The program jumps to this routine when
//       the microcontroller has a power on reset.
//
//******************************************************

void
tristateAll()
{
	port2 = port2_int = HIZPU;	// p2[5:4], XRES = 0
	port1 = port1_int = 0x03;	// p1[1:0]
}

int
main()
{
	// setup data memory stack pointer

	AC = 0x20;
	MOVE_DSP();

	// clear variables      

	ep0_in_machine = EP0_IN_IDLE;
	ep0_transtype = TRANS_NONE;

	configuration = CLEAR;

	suspend_count = 0;
	suspend_flag = CLEAR;

	ep1_stall = CLEAR;
	ep2_stall = CLEAR;

	in_state = 0;
	in_timer = 0;
	in_timeout_value = IN_TIMEOUT;
	bulkMode = 0;
	blockId = 0;
	sequenceMode = 0;
	programMode = 0;

	// configure endpoint usage
	// EP0 @ 0xB8 - 8 bytes
	// EP1 @ 0xC0 - 32 bytes
	// EP2 @ 0xE0 - 32 bytes
	usb_status = ENDPOINT_SIZE;

	// disable HAPI
	hapi_i2c_config = 0;


	// configure GPIO
	gpio_config = (PORT0_CMOS | PORT1_CMOS | PORT2_CMOS | PORT3_CMOS);

	port0 = 0x00;	// Port 0 is used for LED source
	port1 = 0x00;	// Port 1 is I2C / TC
	port2 = 0x00;	// XRES
	port3 = 0x00;	// Port 3 is alternate TC

	// enable global interrupts
	global_int = (_1MS_INT | USB_RESET_INT);

	// enable endpoint0 interrupt
	endpoint_int = EP0_INT;

	tristateAll();

	EI();

	while (1)
	{
		watchdog = AC;

		// SDATA shud be HIZ, in INT mode
		// so we must write SDATA to port1 !

		port0 = 0;	// Turn off USB LED
	}
	return 0;
}

//******************************************************
//
//  Interrupt: Bus_reset
//  Purpose: This routine resets the USB address, 
//          resets the EP0,1,2 modes to their default
//          states, resets the PSP and restarts the
//          program.
//
//******************************************************

void
Bus_reset()
{
//	Clear all of SRAM
#asm
			mov		a,0
			mov		x,0
clearRamLp:	mov	[x+0],a
			inc		x
			jnc		clearRamLp
#endasm

	ep0_mode = STALL_IN_OUT;	// set to STALL INs&OUTs
	ep1_mode = DISABLE;			// disable endpoint1
	ep2_mode = DISABLE;			// disable endpoint2

	usb_address = ADDRESS_ENABLE;	// enable USB address 0

	AC = 0;
	MOVE_PSP();			// reset program stack pointer

	goto main;			// jump to main and restart program
}

//*********************************************************
//
//  Interrupt: 1ms_timer
//  Purpose: This routine keeps track of the USB idle
//          time.  If this interval exceeds 3ms and the
//          device is bus powered, then we must suspend.
//
//*********************************************************

void
_1ms_timer()
{
	// the IN endpoint timeout feature is disabled when the
	// timeout value equals zero.  If this is not the case
	// then we increment the in_timer variable until it equals
	// the in_time_value set by the USB host, or the default
	// set in the hostinf.h file.

	if (in_timeout_value)
	{
		++in_timer;

		// check if timeout has been reached.

		if (in_timer == in_timeout_value)	
		{
			in_state = 0;
			in_timer = 0;
		}
	}
}


//*********************************************************
//************ begin endpoint0 request handlers ***********
//*********************************************************
//
//  Function: Get_status
//  Purpose: This function is called in response to a 
//          GET_STATUS request.  This function responds
//          with the device remote wakeup and power status, 
//          a two byte null packet for interface status, and
//          endpoint stall status.  See section 9.4.5 of
//          the USB 1.1 specification for more details
//
//          DEVICE REQUEST STRUCTURE
//          0x80 0x00 0x00 0x00 0x00 0x00 0x02 0x00
//
//          RESPONSE
//          (000000 REM PWR) 0x00
//
//          INTERFACE REQUEST STRUCTURE
//          0x81 0x00 0x00 0x00 0x00 0x00 0x02 0x00
//
//          RESPONSE
//          0x00 0x00
//
//          ENDPOINT REQUEST STRUCTURE
//          0x82 0x00 0x00 0x00 ENDP ENDP 0x02 0x00
//
//          RESPONSE
//          (0000000 STL) 0x00
//
//
//
//******************************************************

void
Get_status()
{
	if (ep0_dmabuff[BMREQUESTTYPE] == (DEVICETOHOST | DEVICE_RECIPIENT))
	{
		maximum_data_count = DEVICE_STATUS_LENGTH;
		data_start = &device_status_wakeup_disabled;
		Initialize_control_read();
		return;
	}
	else 
	if (ep0_dmabuff[BMREQUESTTYPE] == (DEVICETOHOST | INTERFACE_RECIPIENT))
	{
		maximum_data_count = INTERFACE_STATUS_LENGTH;
		data_start = &interface_status_table;
		Initialize_control_read();
		return;
	}
	else 
	if (ep0_dmabuff[BMREQUESTTYPE] == (DEVICETOHOST | ENDPOINT_RECIPIENT))
	{
		maximum_data_count = ENDPOINT_STALL_LENGTH;

		if (ep0_dmabuff[WINDEXLO] == (IN | 1))
		{
			if (ep1_stall == SET)
				data_start = &endpoint_stall_table;
			else
				data_start = &endpoint_nostall_table;
			Initialize_control_read();
			return;
		}
		else 
		if (ep0_dmabuff[WINDEXLO] == (OUT | 2))
		{
			if (ep2_stall == SET)
				data_start = &endpoint_stall_table;
			else
				data_start = &endpoint_nostall_table;
			Initialize_control_read();
			return;
		}
	}

	Request_not_supported();
}


//******************************************************
//
//  Function: Clear_feature
//  Purpose: This function is called in response to a 
//          CLEAR_FEATURE request.  This function stalls 
//          a "set remote wakeup" request to the device, and
//          clears a stall condition on an endpoint.  See 
//          section 9.4.1 of the USB 1.1 specification for 
//          more information on this request.
//
//          DEVICE REQUEST STRUCTURE
//          0x00 0x01 0x01 0x00 0x00 0x00 0x00 0x00
//
//          ENDPOINT REQUEST STRUCTURE
//          0x02 0x01 0x00 0x00 ENDP ENDP 0x00 0x00
//
//******************************************************

void
Clear_feature()
{
	if (ep0_dmabuff[BMREQUESTTYPE] == (HOSTTODEVICE | ENDPOINT_RECIPIENT))
	{
		if (ep0_dmabuff[WVALUELO] == ENDPOINT_STALL)	// check feature selector
		{
			if (ep0_dmabuff[WINDEXLO] == (IN | 1))	// check endpoint number
			{
				ep1_stall = CLEAR;
				ep1_mode &= ~EP_STALL;
			}
			else if (ep0_dmabuff[WINDEXLO] == (OUT | 2))
			{
				ep2_stall = CLEAR;
				ep2_mode &= ~ACK_OUT;
			}
			else
			{
				Request_not_supported();
				return;
			}

			ep0_in_flag = NO_CHANGE_PENDING;
			Initialize_no_data_control();
			return;
		}
		else
		{
			Request_not_supported();
			return;
		}
	}

	Request_not_supported();
}


//******************************************************
//
//  Function: Set_feature
//  Purpose: This function is called in response to a 
//          SET_FEATURE request.  This function sets
//          a stall condition on an endpoint, and STALLs
//          a request to enable remote wakeup.  See section
//          9.4.9 of the USB 1.1 specification for more
//          information on this request.
//
//          DEVICE REQUEST STRUCTURE
//          0x00 0x03 0x01 0x00 0x00 0x00 0x00 0x00
//
//          ENDPOINT REQUEST STRUCTURE
//          0x02 0x03 0x00 0x00 ENDP ENDP 0x00 0x00
//
//******************************************************

void
Set_feature()
{
	if (ep0_dmabuff[BMREQUESTTYPE] == (HOSTTODEVICE | ENDPOINT_RECIPIENT))
	{
		if (ep0_dmabuff[WVALUELO] == ENDPOINT_STALL)
		{
			if (ep0_dmabuff[WINDEXLO] == (IN | 1))
			{
				ep1_stall = SET;
				ep1_mode |= EP_STALL;
			}
			else 
			if (ep0_dmabuff[WINDEXLO] == (OUT | 2))
			{
				ep2_stall = SET;
				ep2_mode |= EP_STALL;
			}
			else
			{
				Request_not_supported();
				return;
			}
		}
		ep0_in_flag = NO_CHANGE_PENDING;
		Initialize_no_data_control();
		return;
	}

	Request_not_supported();
}


//******************************************************
//
//  Function: Set_address
//  Purpose: This function is called in response to a 
//          SET ADDRESS request.  This function then stores
//          the address and sets a flag, so when the no-data
//          control transaction is complete, the endpoint0
//          IN handler will know to change the device address. 
//
//          REQUEST STRUCTURE
//          0x00 0x05 ADDR 0x00 0x00 0x00 0x00 0x00
//
//******************************************************

void
Set_address()
{
	ep0_in_flag = ADDRESS_CHANGE_PENDING;	// set address change flag
	pending_data = ep0_dmabuff[WVALUELO];	// store address byte

	Initialize_no_data_control();	// Respond with status stage
}


//******************************************************
//
//  Function: Get_descriptor
//  Purpose: This function is called in response to a 
//          GET_DESCRIPTOR request.  This function returns 
//          the requested descriptor.  The valid descriptors
//          include DEVICE, CONFIGURATION, STRING, INTERFACE,
//          and ENDPOINT.
//
//          REQUEST STRUCTURE
//          0x80 0x06 INDX DESC 0x00 0x00 LENG 0x00
//
//          RESPONSE
//          device descriptor
//          configuration descriptor (config, interface, endp1, endp2)
//          string descriptor (product, mfg, config, language)
//          interface descriptor
//          endpoint descriptor (endpoint1 or endpoint2)
//
//
//******************************************************

void
Get_descriptor()
{
	switch (ep0_dmabuff[WVALUEHI])
	{
	case DEVICE:
		{
			maximum_data_count = DEVICE_LENGTH;
			data_start = &device_desc_table;
			Initialize_control_read();
			return;
		}

	case CONFIGURATION:
		{
			maximum_data_count = CONFIG_LENGTH + INTERFACE_LENGTH + 2 * ENDPOINT_LENGTH;
			data_start = &config_desc_table;
			Initialize_control_read();
			return;
		}

	case STRING:
		{
			switch (ep0_dmabuff[WVALUELO])
			{
			case 0x00:
				{
					maximum_data_count = ilanguage_string[0];
					data_start = &ilanguage_string;
					Initialize_control_read();
					return;
				}
			case 0x01:
				{
					maximum_data_count = imanufacturer_string[0];
					data_start = &imanufacturer_string;
					Initialize_control_read();
					return;
				}
			case 0x02:
				{
					maximum_data_count = iproduct_string[0];
					data_start = &iproduct_string;
					Initialize_control_read();
					return;
				}
			default:
				{
					Request_not_supported();
					return;
				}
			}
		}

	case INTERFACE:
		{
			/* NOTE: Chapter 9 of the USB 1.1 specification states */
			/* that the host should never ask for this descriptor  */
			/* by itself */
			Request_not_supported();
			return;
		}

	case ENDPOINT:
		{
			/* NOTE: Chapter 9 of the USB 1.1 specification states */
			/* that the host should never ask for this descriptor  */
			/* by itself */
			Request_not_supported();
			return;
		}
	}

	Request_not_supported();
}


//******************************************************
//
//  Function: Get_configuration
//  Purpose: This function is called in response to a 
//          GET_CONFIGURATION request.  This function returns 
//          the current configuration state.
//
//          REQUEST STRUCTURE
//          0x80 0x08 0x00 0x00 0x00 0x00 0x01 0x00
//
//          RESPONSE
//          0x00 - unconfigured
//          0x01 - configured
//
//******************************************************

void
Get_configuration()
{
	if (configuration == SET)
	{
		maximum_data_count = DEVICE_CONFIG_LENGTH;
		data_start = &device_configured_table;
		Initialize_control_read();
		return;
	}
	else
	{
		maximum_data_count = DEVICE_CONFIG_LENGTH;
		data_start = &device_unconfigured_table;
		Initialize_control_read();
		return;
	}
}


//******************************************************
//
//  Function: Set_configuration
//  Purpose: This function is called in response to a 
//          SET_CONFIGURATION request.  This function then
//          sets the configuration to 0x01 to enable
//          endpoint1,2 or to 0x00 to disable endpoint 1,2. 
//
//          REQUEST STRUCTURE
//          0x00 0x09 CONF 0x00 0x00 0x00 0x00 0x00
//
//******************************************************

void
Set_configuration()
{
	if (ep0_dmabuff[WVALUELO] == 0x01)	// check if device is being configured
	{
		configuration = SET;

		if (ep1_stall == SET)	// set endpoint1 mode
			ep1_mode = ACK_IN | EP_STALL;
		else	ep1_mode = (ep1_mode & EP_STALL) | NAK_IN;


		if (ep2_stall == SET)	// set endpoint2 mode
			ep2_mode = ACK_OUT | EP_STALL;
		else	ep2_mode = (ep2_mode & EP_STALL) | ACK_OUT;

		ep1_data_toggle = 0x00;	// reset the data toggles
		ep2_data_toggle = 0x00;

		endpoint_int = EP0_INT | EP1_INT | EP2_INT;	// enable endpoint0,2 interrupts

		ep0_in_flag = NO_CHANGE_PENDING;
		Initialize_no_data_control();
		return;
	}
	else 
	if (ep0_dmabuff[WVALUELO] == 0x00)	// check if device is being unconfigured
	{
		configuration = CLEAR;

		ep1_mode = DISABLE;
		ep2_mode = DISABLE;

		endpoint_int = EP0_INT;

		ep0_in_flag = NO_CHANGE_PENDING;
		Initialize_no_data_control();
		return;
	}

	Request_not_supported();	// if invalid config value, then STALL request
}


//******************************************************
//
//  Function: Get_interface
//  Purpose: This function is called in response to a 
//          GET_INTERFACE request.  This function returns 
//          the currently selected interface.
//
//          REQUEST STRUCTURE
//          0x81 0x0A 0x00 0x00 INTF INTF 0x01 0x00
//
//          RESPONSE
//          0x00 0x00
//
//******************************************************

void
Get_interface()
{
	maximum_data_count = INTERFACE_ALTERNATE_LENGTH;
	data_start = &interface_alternate_table;
	Initialize_control_read();
}


//******************************************************
//
//  Function: Set_interface
//  Purpose: This function is called in response to a 
//          SET_INTERFACE request.  This function returns 
//          the current configuration.  Only interface 0,
//          with alternate setting 0 will be ACK'd by this
//          function.  Note that this function does not
//          validate WVALUEHI and WINDEXHI values.
//
//          REQUEST STRUCTURE
//          0x01 0x0B ALTR ALTR INTF INTF 0x00 0x00
//
//******************************************************

void
Set_interface()
{
	if ((ep0_dmabuff[WVALUELO] == 0x00) && (ep0_dmabuff[WINDEXLO] == 0x00))
	{
		ep0_in_flag = NO_CHANGE_PENDING;
		Initialize_no_data_control();
		return;
	}

	Request_not_supported();
}

//******************************************************
//
//  Function: Request_not_supported
//  Purpose: This function is called when an invalid or
//          unsupported request is sent from the host.  
//          This function then causes the SETUP transaction
//          to stall.
//
//******************************************************

void
Request_not_supported()
{
	do
	{
		ep0_mode = STALL_IN_OUT;
	}
	while (ep0_mode != STALL_IN_OUT);
}


//*********************************************************
//******** USB vendor specific handling functions *********
//*********************************************************

//	SDATA is in regular CMOS mode

static void
Tx8Bits()
{
	if (programMode)
	{
		if (tcData & 0x80) 
		{
			port3 = S3DATA;
			port3 = S3DATA | S3CLK;
			port3 = S3DATA;
		}
		else 
		{
			port3 = 0;
			port3 = S3CLK;
			port3 = 0;
		}
		if (tcData & 0x40) 
		{
			port3 = S3DATA;
			port3 = S3DATA | S3CLK;
			port3 = S3DATA;
		}
		else 
		{
			port3 = 0;
			port3 = S3CLK;
			port3 = 0;
		}
		if (tcData & 0x20) 
		{
			port3 = S3DATA;
			port3 = S3DATA | S3CLK;
			port3 = S3DATA;
		}
		else 
		{
			port3 = 0;
			port3 = S3CLK;
			port3 = 0;
		}
		if (tcData & 0x10) 
		{
			port3 = S3DATA;
			port3 = S3DATA | S3CLK;
			port3 = S3DATA;
		}
		else 
		{
			port3 = 0;
			port3 = S3CLK;
			port3 = 0;
		}
		if (tcData & 0x08) 
		{
			port3 = S3DATA;
			port3 = S3DATA | S3CLK;
			port3 = S3DATA;
		}
		else 
		{
			port3 = 0;
			port3 = S3CLK;
			port3 = 0;
		}
		if (tcData & 0x04) 
		{
			port3 = S3DATA;
			port3 = S3DATA | S3CLK;
			port3 = S3DATA;
		}
		else 
		{
			port3 = 0;
			port3 = S3CLK;
			port3 = 0;
		}
		if (tcData & 0x02) 
		{
			port3 = S3DATA;
			port3 = S3DATA | S3CLK;
			port3 = S3DATA;
		}
		else 
		{
			port3 = 0;
			port3 = S3CLK;
			port3 = 0;
		}
		if (tcData & 0x01) 
		{
			port3 = S3DATA;
			port3 = S3DATA | S3CLK;
			port3 = S3DATA;
		}
		else 
		{
			port3 = 0;
			port3 = S3CLK;
			port3 = 0;
		}
	}
	else
	{
		if (tcData & 0x80) 
		{
			port1 = SDATA;
			port1 = SDATA | SCLK;
			port1 = SDATA;
		}
		else 
		{
			port1 = 0;
			port1 = SCLK;
			port1 = 0;
		}
		if (tcData & 0x40) 
		{
			port1 = SDATA;
			port1 = SDATA | SCLK;
			port1 = SDATA;
		}
		else 
		{
			port1 = 0;
			port1 = SCLK;
			port1 = 0;
		}
		if (tcData & 0x20) 
		{
			port1 = SDATA;
			port1 = SDATA | SCLK;
			port1 = SDATA;
		}
		else 
		{
			port1 = 0;
			port1 = SCLK;
			port1 = 0;
		}
		if (tcData & 0x10) 
		{
			port1 = SDATA;
			port1 = SDATA | SCLK;
			port1 = SDATA;
		}
		else 
		{
			port1 = 0;
			port1 = SCLK;
			port1 = 0;
		}
		if (tcData & 0x08) 
		{
			port1 = SDATA;
			port1 = SDATA | SCLK;
			port1 = SDATA;
		}
		else 
		{
			port1 = 0;
			port1 = SCLK;
			port1 = 0;
		}
		if (tcData & 0x04) 
		{
			port1 = SDATA;
			port1 = SDATA | SCLK;
			port1 = SDATA;
		}
		else 
		{
			port1 = 0;
			port1 = SCLK;
			port1 = 0;
		}
		if (tcData & 0x02) 
		{
			port1 = SDATA;
			port1 = SDATA | SCLK;
			port1 = SDATA;
		}
		else 
		{
			port1 = 0;
			port1 = SCLK;
			port1 = 0;
		}
		if (tcData & 0x01) 
		{
			port1 = SDATA;
			port1 = SDATA | SCLK;
			port1 = SDATA;
		}
		else 
		{
			port1 = 0;
			port1 = SCLK;
			port1 = 0;
		}
	}
}

void
TxOP()
{
	if (programMode)
	{
		if (tcOp & 0x04) 
		{
			port3 = S3DATA;
			port3 = S3DATA | S3CLK;
			port3 = S3DATA;
		}
		else 
		{
			port3 = 0;
			port3 = S3CLK;
			port3 = 0;
		}
		if (tcOp & 0x02) 
		{
			port3 = S3DATA;
			port3 = S3DATA | S3CLK;
			port3 = S3DATA;
		}
		else 
		{
			port3 = 0;
			port3 = S3CLK;
			port3 = 0;
		}
		if (tcOp & 0x01) 
		{
			port3 = S3DATA;
			port3 = S3DATA | S3CLK;
			port3 = S3DATA;
		}
		else 
		{
			port3 = 0;
			port3 = S3CLK;
			port3 = 0;
		}
	}
	else
	{
		if (tcOp & 0x04) 
		{
			port1 = SDATA;
			port1 = SDATA | SCLK;
			port1 = SDATA;
		}
		else 
		{
			port1 = 0;
			port1 = SCLK;
			port1 = 0;
		}
		if (tcOp & 0x02) 
		{
			port1 = SDATA;
			port1 = SDATA | SCLK;
			port1 = SDATA;
		}
		else 
		{
			port1 = 0;
			port1 = SCLK;
			port1 = 0;
		}
		if (tcOp & 0x01) 
		{
			port1 = SDATA;
			port1 = SDATA | SCLK;
			port1 = SDATA;
		}
		else 
		{
			port1 = 0;
			port1 = SCLK;
			port1 = 0;
		}
	}
}

//	Ouput 3 data 0's

void
Tx30()
{
	// Since SDATA is driven to 0, one can drive SCLK directly

	if (programMode)
	{
		port3 = 0;

		port3 = S3CLK ;
		port3 = 0;

		port3 = S3CLK ;
		port3 = 0;

		port3 = S3CLK ;
		port3 = 0;
	}
	else
	{
		port1 = 0;

		port1 = SCLK ;
		port1 = 0;

		port1 = SCLK ;
		port1 = 0;

		port1 = SCLK ;
		port1 = 0;
	}
}

static void
GrabSdata()
{
	if (programMode)
	{
		port3 = 0;
		port3_int = 0;	// CMOS mode
	}
	else
	{
		port1 = 0;
		port1_int = 0;	// CMOS mode
	}
}

static void
ReleaseBoth()
{
	port1_int = SDATA | SCLK;	// Open Drain when data == 1
	port1 = SDATA | SCLK;		// Data <= 1 -> HiZ
}

static void
ReleaseSdata()
{
	if (programMode)
	{
		port3_int = S3DATA;	// Open Drain when data == 1
		port3 = S3DATA;		// Data <= 1 -> HiZ
	}
	else
	{
		port1_int = SDATA;	// Open Drain when data == 1
		port1 = SDATA;		// Data <= 1 -> HiZ
	}
}

//	In RxByte() the SDATA line if HiZ so SCLK can be driven directly

static void
RxByte()
{
	static	Byte	j;

	if (programMode)
	{
		port3 = 0;
		ReleaseSdata();

		port3 = S3CLK | S3DATA;
		port3 = S3DATA;

		tcData = 0;

		port3 = S3CLK | S3DATA;
		port3 = S3DATA;

		for(j = 9; --j;)
		{
			tcData <<= 1;

			if (port3 & S3DATA) tcData |= 1;

			port3 = S3CLK | S3DATA;
			port3 = S3DATA;
		}

		port3 = S3CLK | S3DATA;
		port3 = S3DATA;
	}
	else
	{
		port1 = 0;
		ReleaseSdata();

		port1 = SCLK | SDATA;
		port1 = SDATA;

		tcData = 0;

		port1 = SCLK | SDATA;
		port1 = SDATA;

		for(j = 9; --j;)
		{
			tcData <<= 1;

			if (port1 & SDATA) tcData |= 1;

			port1 = SCLK | SDATA;
			port1 = SDATA;
		}

		port1 = SCLK | SDATA;
		port1 = SDATA;
	}
}

// Uses globals tcOp, addr, and data

void
TestController()
{
	GrabSdata();

	if (tcOp == TCOP_IOR)
	{
		TxOP();
		tcData = addr; Tx8Bits();
		RxByte();
	}
	else
	if (tcOp == TCOP_MR)
	{
		TxOP();
		tcData = addr; Tx8Bits();
		RxByte();
	}
	else
	if (tcOp == TCOP_IOW)
	{
		TxOP();
		tcData = addr; Tx8Bits();
		tcData = data; Tx8Bits();
		Tx30();
	}
	else
	if (tcOp == TCOP_MW)
	{
		TxOP();
		tcData = addr; Tx8Bits();
		tcData = data; Tx8Bits();
		Tx30();
	}

	// Assume SDATA is Resistive 0 from PSOC
	// Idle the SDATA in HiZ from USB

	ReleaseSdata();
	watchdog = AC;
}

void
delay50us()
{
	for(delay1 = 48; --delay1; ) ;
	watchdog = AC;
}

void
delay5us()
{
	for(delay1 = 2; --delay1; ) watchdog = AC;
}

static void
PSOCreset()
{
	port2 = XRES | HIZPU;
	delay50us();
	port2 = HIZPU;
	delay5us();
}

static void
Acquire()
{
	tristateAll();
	PSOCreset();
	GrabSdata();

	if (programMode)
	{
		port3 = S3CLK ;	// Clock a single 0
		port3 = 0x00 ;
	}
	else
	{
		port1 = SCLK ;	// Clock a single 0
		port1 = 0x00 ;
	}

	blockId = 0;
	sequenceMode = 0;
	bulkMode = 0;

	tcData = 0x0CA;
	Tx8Bits();

	// Send 5 bytes of 0s

	tcData = 0;
	Tx8Bits();
	Tx8Bits();
	Tx8Bits();
	Tx8Bits();
	Tx8Bits();

	// PSOC SDATA shud remain in Resistive 0 for 16ms
	// Tri-stating SDATA here should not be a problem

	tcOp = TCOP_IOR;
	addr = 0xFF;
	// data = X;
	TestController();

	data = tcData;
}

//******************************************************
//
//  Function: regWrite
//  Purpose: This function is called in response to a 
//          REGWRITE request.  The regWrite command is 
//          then written into the PLD along with the address
//          and data for the PLD register.
//
//  REQUEST STRUCTURE
//  0x40 0x00 {0x00 0x00} {0x00 0x00} {0x00 0x00}
//  IN    ID   Cmd  Mode   Data xxxx   Count
//******************************************************

void
regWrite()
{
	ep0_mode &= 0x7f;
	
	cmd  = ep0_dmabuff[2];
	mode = ep0_dmabuff[3];
	data = ep0_dmabuff[4];

	if (cmd == SRAM)
	{
		sram_address = data;
	}
	else
	if (cmd == BULKMODES)
	{
		bulkMode = mode;
		blockId = 0;
		sequenceMode = 0;
	}
	else
	if (cmd == DUT_MEM)
	{
		tcOp = TCOP_MW;
		addr = mode;
		// data = data
		TestController();	// RAM, ADDR, DATA
	}
	else
	if (cmd == DUT_IO)
	{
		tcOp = TCOP_IOW;
		addr = mode;
		// data = data
		TestController();	// IO, ADDR, DATA
	}
	else 
	if (cmd == PSOC)
	{
		if (mode == PSOC_I2C)
		{
			// Ensure I2C Pullups enabled
			port2 = HIZPU;
			port2_int = 0;

			// Ensure I2C bus is idle
			port1 = port1_int = SCLK | SDATA;

			addr  = ep0_dmabuff[2];
			data = ep0_dmabuff[3];	// command or address

			PSOCreset();	// Pulse XRES, keeping bus idle
		}
		else
		if (mode == PSOC_RESET)
		{
			tristateAll();
			PSOCreset();
		}
		else
		if (mode == PSOC_KEY)
		{
			if (data) programMode = 0;
			else 
			{
				programMode = 1;
				ReleaseBoth();	// Dont use Port1
			}

			Acquire();
		}
	}

	ep0_in_flag = NO_CHANGE_PENDING;
	Initialize_no_data_control();
}

void
ep0DataReturn(Byte count)
{
	maximum_data_count = 0;

	do
	{
		ep0_transtype = TRANS_CONTROL_READ;	// set current transaction to control read
	}
	while (ep0_transtype != TRANS_CONTROL_READ);

	count |= DATA_TOGGLE;

	do
	{
		ep0_count = count;		// setup number of bytes to send
	}
	while (ep0_count != count);

	do
	{
		ep0_mode = ACK_IN_STATUS_OUT;	// set send mode    
	}
	while (ep0_mode != ACK_IN_STATUS_OUT);
}

//******************************************************
//
//  Function: regRead
//  Purpose: This function is called in response to a 
//          regRead request.  The regRead command is 
//          then written into the PLD.  The PLD then knows
//          to return the contents of the register to 
//          the CY7C64013.
//
//  REQUEST STRUCTURE
//  0xC0 0x01 {0x00 0x00} {0x00 0x00} {0x01 0x00}
//  OUT  ID    Cmd  Mode   xxxx xxxx   COUNT
//******************************************************

void
regRead()
{
	ep0_mode &= 0x7f;

	/* write out read command to the PLD */

	cmd  = ep0_dmabuff[2];
	mode = ep0_dmabuff[3];	// command or address

	if (cmd == DUT_MEM)
	{
		tcOp = TCOP_MR;
		addr = mode;
		// data = X;
		TestController();
		data = tcData;
	}
	else
	if (cmd == DUT_IO)
	{
		tcOp = TCOP_IOR;
		addr = mode;
		// data = X;
		TestController();
		data = tcData;
	}
	else
	if (cmd == PSOC)
	{
		if (mode == PSOC_POLL)
		{
			data = 0;
			if (programMode)
			{
				if (port3 & S3DATA) data = 1;

				port3 = S3CLK | S3DATA;
				port3 = S3DATA;
			}
			else
			{
				if (port1 & SDATA) data = 1;

				port1 = SCLK | SDATA;
				port1 = SDATA;
			}
		}
	}

	ep0_dmabuff[0] = data;

	/* send the data back up endpoint 0 */

	ep0DataReturn(1);
}

//******************************************************
//
//  Function: i2cWrite
//  Purpose: This function is called in response to a
//          I2CWRITE request.
//
//  REQUEST STRUCTURE
//  0x40 0x02 {0x00 0x00}   {0x00  0x00} {0x00 0x00}
//  IN   ID    ADDR OFFSET   D0    D1     xxxx xxxx
//******************************************************

// Offset [7:6] contains count

void
i2cWrite()
{
    ep0_mode &= 0x7f;

    // Ensure I2C Pullups enabled
    port2 = HIZPU;
    port2_int = 0;

    // Ensure I2C bus is idle
    port1 = port1_int = SCLK | SDATA;

    I2CStart();
    tcData = ep0_dmabuff[2];    // Addar

    if (I2CTx())    // NAK
    {
        return;
    }

    count = (ep0_dmabuff[3] & 0xC0) >> 6;
    
    tcData = ep0_dmabuff[3] & 0x03F;// Offset

    (void) I2CTx();

    if (count)
    {
        tcData = ep0_dmabuff[4];

        (void) I2CTx();

        if (count > 1)
        {
            tcData = ep0_dmabuff[5];

            (void) I2CTx();
        }
    }
    I2CStop();
}


//******************************************************
//
//  Function: i2cRead
//  Purpose: This function is called in response to a
//          I2CREAD request.
//
//  REQUEST STRUCTURE
//  0x40 0x03 {0x00 0x00}   {0x00  0x00} {0x00 0x00}
//  IN    ID   ADDR Count    xxxx  xxxx   xxxx
//******************************************************


void
i2cRead()
{
    ep0_mode &= 0x7f;

    // Ensure I2C Pullups enabled
    port2 = HIZPU;
    port2_int = 0;

    // Ensure I2C bus is idle
    port1 = port1_int = SCLK | SDATA;

    count = ep0_dmabuff[3];

    if (count > 8) return;

    I2CStart();
    tcData = ep0_dmabuff[2];    // Address

    if (I2CTx())
    {
        ep0DataReturn(0);       // send data back

        return;
    }

    for(j = 0; j < count; )
    {
        ++j;

        if (j == count) I2CRx(1);   // NAK last byte
        else I2CRx(0);

        ep0_dmabuff[j-1] = tcData;
    }

    I2CStop();

    ep0DataReturn(count);       // send data back
}

//*********************************************************
//********** USB transaction handling functions ***********
//*********************************************************


//******************************************************
//
//  Function: Initialize_control_read
//  Purpose: This function is called when a SETUP packet
//          has been successfully parsed and we are ready
//          to respond with a control read transaction.
//          First we must determine the size of the response,
//          which is the lesser of the requested size and
//          the total size of the response table.  We also
//          have to set the correct data toggle and begin
//          writing bytes to the endpoint fifo.
//
//******************************************************

void
Initialize_control_read()
{

	ep0_transtype = TRANS_CONTROL_READ;	// set current transaction to control read
	ep0_data_toggle = DATA_TOGGLE;	// set to DATA1, which should be the data

	// toggle of the resposne to an IN

	if (ep0_dmabuff[WLENGTHHI] == 0)
	{
		if (ep0_dmabuff[WLENGTHLO] < maximum_data_count)
			maximum_data_count = ep0_dmabuff[WLENGTHLO];
	}

	Control_read_data_stage();
}


//******************************************************
//
//  Function: Initialize_control_write
//  Purpose: This function is called when a SETUP packet
//          has been successfully parsed and we are ready
//          to accept a control write transaction.  Currently
//          this functionality is unsupported, but the basic
//          framework is here.
//
//******************************************************

void
Initialize_control_write()
{
	ep0_transtype = TRANS_CONTROL_WRITE;
	ep0_data_toggle = DATA_TOGGLE;

	do
	{
		ep0_mode = ACK_OUT_STATUS_IN;
	}
	while (ep0_mode != ACK_OUT_STATUS_IN);

}


//******************************************************
//
//  Function: Initialize_no_data_control
//  Purpose: This function is called when a SETUP packet
//          has been successfully parsed and we are ready
//          to respond with the status stage of a no-data
//          control transaction.
//
//******************************************************

void
Initialize_no_data_control()
{
	ep0_transtype = TRANS_NO_DATA_CONTROL;
	do
	{
		ep0_mode = STATUS_IN_ONLY;
	}
	while (ep0_mode != STATUS_IN_ONLY);
}


//******** End USB transaction handling functions *********
//*********************************************************
//*********************************************************

//*********************************************************
//*********************************************************
//*********** Response to an endpoint0 IN *****************

//******************************************************
//
//  Function: Control_read_data_stage
//  Purpose: This function is called when a IN token has
//          been received on endpoint0 and the transaction
//          type is selected to be TRANS_CONTROL_READ.  This
//          function then loads the data buffer with the
//          current 
//
//******************************************************

void
Control_read_data_stage()
{
	int_temp = 0;

	ep0_mode = NAK_IN_OUT;		// clear register in order to enable
	// writes to the endpoint FIFO

	while (maximum_data_count != 0 && int_temp != 8)
	{
		ep0_dmabuff[int_temp] = *data_start;	// write to the FIFO

		++data_start;
		++int_temp;
		--maximum_data_count;
	}

	AC = ep0_count;

	do
	{
		ep0_count = ep0_data_toggle | int_temp;	// setup number of bytes to send
	}
	while (ep0_count != (ep0_data_toggle | int_temp));

	do
	{
		ep0_mode = ACK_IN_STATUS_OUT;	// set send mode    
	}
	while (ep0_mode != ACK_IN_STATUS_OUT);

	ep0_data_toggle ^= DATA_TOGGLE;	// change data toggle
}


//******************************************************
//
//  Function: Control_write_status_stage
//  Purpose: This function is called when a IN token has
//          been received on endpoint0 and the transaction
//          type is selected to be TRANS_CONTROL_WRITE. The 
//          IN is the status stage which ends the transaction 
//
//******************************************************

void
Control_write_status_stage()
{
	do
	{
		ep0_mode = STATUS_IN_ONLY;
	}
	while (ep0_mode != STATUS_IN_ONLY);

	ep0_transtype = TRANS_NONE;
}

//******************************************************
//
//  Function: no_data_control_status_stage
//  Purpose: This function is called when a IN token has
//          been received on endpoint0 and the transaction
//          type is selected to be TRANS_NO_DATA_CONTROL.  
//          IN is the status stage which ends the transaction.
//          If the request was an USB address change, then we
//          must change the device address. 
//
//******************************************************

void
No_data_control_status_stage()
{
	if (ep0_in_flag == ADDRESS_CHANGE_PENDING)
		usb_address = pending_data | ADDRESS_ENABLE;

	do
	{
		ep0_mode = STATUS_IN_ONLY;
	}
	while (ep0_mode != STATUS_IN_ONLY);

	ep0_transtype = TRANS_NONE;
}


//*********************************************************
//********** End Endpoint 0 Response to an IN *************
//*********************************************************

//*********************************************************
//*********** Response to an endpoint0 OUT ****************
//*********************************************************

//******************************************************
//
//  Function: Control_read_status_stage
//  Purpose: This function is called when a OUT token has
//          been received on endpoint0 and the transaction
//          type is selected to be TRANS_CONTROL_READ.  This
//          function is called in response to a control read
//          status stage. 
//
//******************************************************

void
Control_read_status_stage()
{
	do
	{
		ep0_mode = STATUS_OUT_ONLY;
	}
	while (ep0_mode != STATUS_OUT_ONLY);

	ep0_transtype = TRANS_NONE;
}


//******************************************************
//
//  Function: Control_write_data_stage
//  Purpose: This function is called when a OUT token has
//          been received on endpoint0 and the transaction
//          type is selected to be TRANS_CONTROL_WRITE.  This
//          function gathers the data sent by the host.
//          NOTE: this function is currently not used and
//          so is not fully implemented for any function. 
//
//******************************************************

void
Control_write_data_stage()
{
	if (ep0_count & DATA_VALID)
	{
		if (ep0_count & DATA_TOGGLE == ep0_data_toggle)
		{
			//get data here

			ep0_data_toggle ^= DATA_TOGGLE;
			ep0_mode = ACK_OUT_STATUS_IN;
			return;
		}
	}

	do
	{
		ep0_mode = NAK_OUT_STATUS_IN;
	}
	while (ep0_mode != NAK_OUT_STATUS_IN);
}


//******************************************************
//
//  Function: No_data_control_error
//  Purpose: This function is called when a OUT token has
//          been received on endpoint0 and the transaction
//          type is selected to be TRANS_NO_DATA_CONTROL.
//          If this happens, then there has been an error
//          somewhere in the transaction process. 
//
//******************************************************

void
No_data_control_error()
{
	do
	{
		ep0_mode = STALL_IN_OUT;
	}
	while (ep0_mode != STALL_IN_OUT);

	ep0_transtype = TRANS_NONE;
}

//************* End Endpoint 0 handling ****************
//************* begin endpoint0 request parsing ***********

const int (*bRequest_jumptable[]) () =
{
	Get_status, Clear_feature, Request_not_supported,	// Reserved for future use
	Set_feature, Request_not_supported,	// Reserved for future use
	Set_address, Get_descriptor, Request_not_supported,	// Set Descriptor not supported
	Get_configuration, Set_configuration, Get_interface, Set_interface, 
	Request_not_supported	// Synch Frame not supported
};

const int (*vendor_jumptable[]) () = 
{
	regWrite, 
	regRead,
	i2cWrite,
	i2cRead
};

const int (*ep0_in_jumptable[]) () =
{
	Request_not_supported,
	Control_read_data_stage,
	Control_write_status_stage,
	No_data_control_status_stage
};

const int (*ep0_out_jumptable[]) () =
{
	Request_not_supported,
	Control_read_status_stage,
	Control_write_data_stage, No_data_control_error
};


//******************************************************
//
//  Interrupt: endpoint0
//  Purpose: USB control endpoint.  This interrupt
//          handler formulates responses to SETUP and 
//          CONTROL READ, and NO-DATA CONTROL transactions
//          on endpoint0. 
//
//******************************************************

static Byte rid;

void
Endpoint0()
{
	// if it's a SETUP token, then parse 
	// it and handle the request

	port0 = LED;

	if (ep0_mode & EP0_SETUP)
	{
		if ((ep0_dmabuff[BMREQUESTTYPE] & 0x60) == STANDARD_TYPE)	// test if standard type
		{
			rid = ep0_dmabuff[BREQUEST];
			(*bRequest_jumptable[rid])();
		}
		else 
		if ((ep0_dmabuff[BMREQUESTTYPE] & 0x60) == VENDOR_TYPE)
		{
			rid = ep0_dmabuff[BREQUEST];

			if (rid < 4) (*vendor_jumptable[rid])();
			else Request_not_supported();
		}
		else
		{
			Request_not_supported();
		}
	}

	// If it's an IN token, then determine 
	// the type of transaction and respond
	// If it's an OUT token, then determine
	// the type of transaction and respond

	else 
	if (ep0_mode & EP0_IN)
	{
	        (*ep0_in_jumptable[ep0_transtype])();	// jump to IN handler
	}
	else 
	if (ep0_mode & EP0_OUT)
	{
	        (*ep0_out_jumptable[ep0_transtype])();	// jump to OUT handler
	}
}

void
Sclk32()
{
	// Assumption is that we are NOT driving SDATA
	// port1 is in CMOS INT mode for SDATA
	// Therefore SDATA shud be written as a 1

	for(i = 33; --i; )
	{
		port1 = SCLK | SDATA;
		port1 = SDATA;
	}
}

void
shortSequence()
{
	tcOp = TCOP_MR;

	for (i = 0; i < 32; ++i)
	{
		addr = sram_address++;
		TestController();

		ep1_dmabuff[i] = tcData;
	}
}

//******************************************************
//
//  Interrupt: Endpoint1
//  Purpose: This endpoint handles sending data from the
//          PLD during an SRAM read to the USB host via IN
//          transactions.
//
//******************************************************

void
Endpoint1()
{
	port0 = LED;
	ReleaseSdata();

	// load the endpoint fifos

	if (!(ep1_mode & EP_ACK))	// check if EP ACK'd or NAK'd
	{
		if ((ep1_mode & 0x0f) != ACK_IN)
		{
			shortSequence();

			// send the data back up endpoint 1
			//maximum_data_count = 0;

			ep1_count = ep1_data_toggle | 32;
			ep1_mode = ACK_IN;	// set send mode    
	
			ep1_data_toggle ^= DATA_TOGGLE;	// change the data toggle
		}
	}
	else
	{
		ep1_mode = NAK_IN;
	}
}

//******************************************************
//
//  Interrupt handler: endpoint2
//  Purpose: This endpoint handles the reception of data
//          from the USB host.  Further data will not be
//          accepted until the current data is read by
//          the PLD.
//
//******************************************************

void
Endpoint2()
{
	static Byte cnt, i;

	port0 = LED;
	ReleaseSdata();

	if (ep2_mode & EP_ACK)
	{
		if ((ep2_count & DATA_TOGGLE) == ep2_data_toggle)	// check the data toggle
		{
			cnt = (ep2_count & BYTE_COUNT) - 2;

			if (bulkMode == QUE) Que();
			else
			for (i = 0; i < cnt; ++i)
			{
				tcOp = TCOP_MW;
				addr = sram_address++;
				data = ep2_dmabuff[i];
				TestController();
			}
			ep2_data_toggle ^= DATA_TOGGLE;	// switch data toggles
		}
	}
	ep2_mode = ACK_OUT;
}

void
Que()
{
	static	Byte i;

	addr = 0xf8;
	tcOp = TCOP_MW;

	// 8 bytes of RAM starting at F8
	for(i = 0; i < 8; ++i)
	{
		data = ep2_dmabuff[i];
		TestController();
		++addr;
	}

	addr = 0x0F0;
	tcOp = TCOP_IOW;

	// 16 bytes of IO starting at F0, continue with i=8
	for(;i < 24; ++i)
	{
		data = ep2_dmabuff[i];
		TestController();
		++addr;
	}
	bulkMode = 0;	// reset back to default

	ReleaseSdata();

	// Give enough clocks to start operation, so Polling works
	Sclk32();
}


//  Both CLK and DATA are 1s going into I2CStart()
//  Drop DATA, CLK will drop in next BYTE operation


static void
I2CStart()
{
    port1 = SCLK;   // SDATA -> 0, CLK still high

    delay5us();
    delay5us();
}


//  CLK is 1, data is 1/0
//  Drop CLK, DROP DATA

static void
I2CStop()
{
    // Drop clock, but maintain Thd

    if (port1 & SDATA)
    {
        port1 = SDATA;  // Drop CLK, DATA remains 1
    }

    // IF DATA was 0 just drop CLK;
    // IF DATA was 1 CLK was droped above, and now drop DATA

    port1 = 0;  // Drop CLK, DATA drive DATA to 0

    delay5us();

    port1 = SCLK;   // Raise CLK, keeping DATA low


    // Wait for SCLK release to rise
    while(0 == (port1 & SCLK));


    // Drive DATA high during CLK high -> STOP

    port1 = SCLK | SDATA;

    delay5us();
}

static void
I2CBit(Byte d)
{

    if (port1 & SDATA)
    {
        // DATA is high and must stay high till after CLK falls

        port1 = SDATA;  // Drive CLK low, leave data valid

        if (!d) port1 = 0;
    }
    else
    {
        // SDATA needs to stay low till after CLK falls

        port1 = 0;

        if (d) port1 = SDATA;
    }

    // DATA is now valid and CLK is low

    delay5us();

    // Now raise CLK, keeping DATA valid

    if (d) port1 = SDATA | SCLK;
    else port1 = SCLK;

    // Wait for SCLK release to rise
    while(0 == (port1 & SCLK));

    delay5us();
}

// Pullups are enabled on Port2
// Data and Clock are the same as test controller
// Bus is idling with pullups at 1s


static int
I2CTx()
{
    // Send the 8 bits of tcData MSB first
    // Requires SCLK = 0 on entry
    // If ACK'd then SCLK = 0 on exit, else bus is idle

    for(i = 9; --i;)
    {
        if (tcData & 0x80) I2CBit(1);
        else I2CBit(0);

        tcData <<= 1;
    }

    // Now do ACK/NACK cycle

    // Pullup state on DATA, SCLK->0, Ensure Thd

    if (port1 & SDATA)
    {
        port1 = SDATA;  // Leave DATA high, drop CLK
    }
    else
    {
        port1 = 0;

        port1 = SDATA;  // Now go into pullup mode
    }

    delay5us();

    // Pullup state on DATA, SCLK->1
    port1  = SDATA | SCLK;

    // Wait for SCLK release to rise
    while(0 == (port1 & SCLK));

    delay5us();

    if (port1 & SDATA)  // NACK
    {
        // SCLK and SDATA == 1, idle
        return 1;
    }
    else
    {
        // SDATA == 0, SCLK == 1
        watchdog = AC;

        return 0;
    }

    // CLK is 1, DATA is ACK state (0, or NAK 1)
}

static void
I2CRx(Byte nack)
{

    // Receive the 8 bits into tcData MSB first
    // Drives SCLK = 0 on entry, SCLK = 1 on exit
    // SDATA = 1 (pullup) thru-out until ACK/NACK

    for(i = 9; --i;)
    {
        tcData <<= 1;

        port1 = SDATA;              // Pullup, and CLK =0

        delay5us();

        port1 = SDATA | SCLK;       // Pullup, try to raise CLK

        while(0 == (port1 & SCLK)); // Wait for SCLK +

        if (port1 & SDATA) tcData |= 1;

        delay5us();
    }

    // Send ACK/NACK to PSoC

    if (port1 & SDATA)
    {
        port1 = SDATA;          // Keep DATA high, Drop CLK

        if (!nack) port1 = 0;       // CLK is low, now drop DATA
    }
    else
    {
        // DATA is low

        port1 = 0;          // Drop CLK, keeping DATA constant

        if (nack) port1 = SDATA;    // Raise DATA, keeping CLK low
    }

    delay5us();

    if (nack)
    {
        port1  = SDATA | SCLK;      // Pulllup, CLK +
    }
    else
    {
        port1  = SCLK;          // 0, CLK +
    }

    // Wait for SCLK release to rise
    while(0 == (port1 & SCLK));

    delay5us();

    // CLK is free to drop in next frame
}

